#include <Wire.h>
#include <AltSoftSerial.h>
#include <util/delay.h>
#include <limits.h>
#include <Arduino.h>
#include <avr/sleep.h>
#include <avr/eeprom.h>
#include <string.h>
#include <ctype.h>


//стрингификатор
#define XSTR(S)		STR(S)
#define STR(S)		#S

//номера консолей
#define TTY_BT		1
#define TTY_SERIAL	2
#define TTY_ALL		3

#define LOWBAT_LEVEL        3000	//минимально допустимое напряжение питания (из-за BT модуля)
#define TTY_BUF_SIZE        96		//размер кольцевого буфера для входящих комманд
#define MAX_TOKEN_SIZE      64		//максимальный размер токена в команде языка управления
#define ENDL                "\r\n"	//символы завершения строки терминала TTY
#define BT_TTY_SPEED_SLAVE  9600	//скорость обмена с BT в режиме SLAVE
#define BT_TTY_SPEED_MASTER 38400	//скорость обмена с BT в режиме MASTER
#define BT_TTY_SPEED_AT     38400	//скорость обмена с BT в режиме AT комманд

//распиновка выводов МК
#define PIN_LED			10     // LED connected to digital pin DIP16
#define PIN_PHOTOR		2      // порт фоторезистора (INT0)
#define PIN_GERKON1		3      // пин первого геркона (INT1)
#define PIN_GERKON2		14     // пин второго геркона (PCINT8)
//#define PIN_RXSOFTTTY		8      // пин RX software tty (PCINT0) НЕ ПЕРЕОПРЕДЕЛЯЕТСЯ
//#define PIN_TXSOFTTTY		X      // пин TX software tty (PCINT1) НЕ ПЕРЕОПРЕДЕЛЯЕТСЯ
#define PIN_BTTX		0      // пин TX на BTмодуле (RX пин МК) (направление BT->МК) (вход)
#define PIN_BTRX		1      // пин RX на BTмодуле (TX пин МК) (направление МК->BT) (выход)
#define PIN_BTPOWER		15     // пин управление питанием BT модуля (PC1) (выход)
#define PIN_BTLED		16     // пин LED на BTмодуле (PC2) (направление BT->МК) вход
#define PIN_BTKEY		17     // пин KEY на BT модуле (PC3) (направление МК->BT) (выход)

//в этой структуре данные сохраняются в eeprom при выполнении
//команд SAVEEEPROM и REBOOT
typedef struct
{
    uint16_t imp01KWH;		// импульсов на 0.1 киловатчас
    // счетчики измеряемых величин
    uint32_t value[3];		//эл-во 0.1 киловатчасы (1235 соотв 123.5kwh), холодная вода литры, горячая вода литры
    uint16_t counter[3];	//импульсы с датчиков
    uint8_t  gap[3];		//определяет насколько часто нужно передавать показания по каждому датчику
    // счетчики наработки (нужны только для отчета юзеру)
    uint32_t cpuuptimems;       //общее время работы CPU милисек (включая время в SLEEP_MODE_IDLE, но без учета SLEEP_MODE_PWR_DOWN)
    uint32_t btuptimems;        //общее время работы BT милисек
    uint8_t  crc;		//контрольная сумма
} State;

State                  state; //текущие величины сохраняемые в eeprom (не теряются при перезагрузках)
State EEMEM            eepromstate; //адрес state в eeprom


// копия счетчиков counter[0], value[0]-value[2]
uint16_t      c0;
uint32_t      v[3];

volatile uint16_t      imp10L           = 2;      // импульсов на 10 литров
volatile bool          ledenable        = true;  //светодиодный индикатор вкл/выкл
volatile bool          uartenable       = false; //true если аппаратный uart включен
volatile bool          btpoweron        = false; //true если включено питание BT
volatile bool          btcmdmode        = false; //true если BT находится в режиме AT команд
volatile bool          btmastermode     = true;  //true если BT находится в master режиме
volatile bool          btcmodezero      = false; //true если BT находится в CMODE=0 режиме
volatile uint32_t      btrxbytes        = 0;    //сколько байт принято за текущий сеанс от BT
volatile uint32_t      btsessionok      = 0;  //счетчик успешных BT сессий
volatile uint32_t      btsessionerr     = 0;  //счетчик неуспешных BT сессий
volatile unsigned long btstarttime      = 0; //время последнего включения BT
unsigned long          uptimeresettime  = 0; //время сброса счетчиков аптайма
volatile bool          sleeping         = false; //флаг устанавливается перед засыпанием
volatile bool          ttywakeup        = false; //флаг устанавливается после пробуждения от консолей

// буфер для клавиатурных символов приходящих с консолей (bt или uart)
uint8_t ttybuf[TTY_BUF_SIZE];    //когда буфер пуст хвост == голове
uint8_t ttybufhead = 0;          //указатель на голову буфера (самый старый эл-т)
uint8_t ttybuftail = 0;          //указатель на хвост (индекс, следующей свободной ячейки)

AltSoftSerial ttySerial(-1,-1); //RX PB0 (INPUT_PULLUP 14pinDIP) TX PB1 (OUTPUT 15pinDIP)
void(* resetFunc) (void) = 0; // Reset MC function

// декларации функций
bool execAtCmd(const char* cmd, uint16_t timeoutsec); //диалог с BT модулем
void serialprintf(uint8_t tty, const char *format,...); //выводит строку на консоль


// *****************************************************************************
// * ОБРАБОТЧИКИ ПРЕРЫВАНИЙ
// *****************************************************************************

//ISR(int0) //обработчик прерывания от фоторезистора
void inthandlerPhoto() //обработчик прерывания от фоторезистора
{
    state.counter[0]++;
    if (state.counter[0] >= state.imp01KWH) //насчитали 0.1 киловат
    {
        state.counter[0] = state.counter[0] - state.imp01KWH;
        state.value[0]++; //увеличиваем счетчик 0.1 киловат
    }
}


//ISR(int1) //обработчик прерывания от gerkon1
void inthandlerGerkon1() //обработчик прерывания от gerkon1
{
    state.counter[1]++;
    if (state.counter[1] >= imp10L) //насчитали литр
    {
        state.counter[1] = state.counter[1] - imp10L;
        state.value[1] += 10; //увеличиваем счетчик литров
    }
}


ISR(PCINT1_vect) //обработчик прерывания от gerkon2
{
    state.counter[2]++;
    if (state.counter[2] >= imp10L) //насчитали литр
    {
        state.counter[2] = state.counter[2] - imp10L;
        state.value[2] += 10; //увеличиваем счетчик литров
    }
}


ISR(PCINT0_vect) //приход симвода на ALT SOFT TTY
{
    if (sleeping)
	ttywakeup = true; //даем основному циклу знать что проснулсь от tty
}


ISR(PCINT2_vect) //обработчик прерывания от hardware tty RX
{
    if (sleeping)
	ttywakeup = true; //даем основному циклу знать что проснулсь от uart
}


// *****************************************************************************
// * ВСЯКИЕ РАЗНЫЕ ФУНКЦИИ
// *****************************************************************************


int freeRam ()
{
    extern int __heap_start, *__brkval;
    int v;
    return (int) &v - (__brkval == 0 ? (int) &__heap_start : (int) __brkval);
}


uint8_t crc8(uint8_t* p, uint16_t size)
{
    uint8_t result = 0;
    for (uint16_t i = 0; i < size; i++)
    {
	result += p[i];
    }
    return result;
}


// *****************************************************************************
// * КОЛЬЦЕВОЙ БУФЕР КЛАВИАТУРНЫХ КОММАНД
// *****************************************************************************


void ttyBufPush(uint8_t ch)
{
    ttybuf[ttybuftail] = ch;
    ttybuftail++;
    if (ttybuftail == TTY_BUF_SIZE) //хвост уперся в конец
	ttybuftail = 0;
    if (ttybuftail == ttybufhead) //хвост догнал голову
    {
	ttybufhead++; //сдвигаем голову
        if (ttybufhead == TTY_BUF_SIZE) //голова уперлась в конец
	    ttybufhead = 0;
    }
}


uint8_t ttyBufGet() //вернет указатель на самый старый эл-т буфера (или 0 если буфер пуст)
{
    uint8_t result = 0;
    if (ttybufhead != ttybuftail)
    {
	result = ttybuf[ttybufhead];
    }
    return result;
}


void ttyBufPop() //удалить самый старый элемент из списка
{
    if (ttybufhead != ttybuftail)
	ttybufhead++;
    if (ttybufhead == TTY_BUF_SIZE) //голова уперлась в конец
	ttybufhead = 0;
}


void ttyBufPopBack() //удалить самый последний элемент из списка
{
    if (ttybufhead == ttybuftail)
	return; //буфер пуст
    if (ttybuftail == 0) //ttybuftail указывает начало буфера
	ttybuftail = TTY_BUF_SIZE-1; //перепрыгнуть в конец
    else
	ttybuftail--; //просо сдвигаемся влево
}


void ttyBufClear() //очистить буфер
{
    ttybufhead = ttybuftail;
}


uint8_t* getTokenFromTtyBuf(uint8_t* buf, int bufsize) //получить токен из буфера или "" память должна быть выделена заранее
{
    uint8_t n = 0;
    uint8_t ch;
    while (true)
    {
	ch = ttyBufGet();
	ttyBufPop();
	buf[n] = ch;
	if ((ch == ' ')||(ch == '\t')||(ch == 0))
	{
	    buf[n]='\0';
	    break;
	}
	if (n < bufsize - 1)
	     n++;
	else //переполнение буфера
	{
	    buf[n]='\0';
	    break;
	}
	
    }
    //serialprintf(false,"{%s}\n\r", buf);
    return buf;
}


// *****************************************************************************
// * ФУНКЦИИ РАБОТЫ С КОНСОЛЯМИ
// *****************************************************************************


void serialprintf(uint8_t tty, const char *format,...) //выводит строку на консоль
{
    char buff[128];
    va_list args;
    va_start (args,format);
    vsnprintf(buff,sizeof(buff),format,args);
    va_end (args);
    buff[sizeof(buff)/sizeof(buff[0])-1]='\0';
    if ((TTY_ALL == tty)||(TTY_BT == tty))
    {
	if (uartenable)
	    Serial.print(buff); //вывод на BT модуль
    }
    if ((TTY_ALL == tty)||(TTY_SERIAL == tty))
	ttySerial.print(buff); //вывод на софтовую tty консоль
}


int ttyavailable(uint8_t tty) //число байт ожидающих прочтения на консоле
{
    if (TTY_BT == tty)
    {
	if (uartenable)
	    return Serial.available();
    }
    else
	return ttySerial.available();
    return 0;
}


uint8_t ttyread(uint8_t tty) //прочитать символ с консоли
{
    if (TTY_BT == tty)
    {
	if (uartenable)
	{
	    uint8_t result = Serial.read();
	    //serialprintf(false,"\n{%0X}\n",result);
	    btrxbytes++;
	    return result;
	}
    }
    else
	return ttySerial.read();
    return ' ';
}


bool ttyenable(uint8_t tty) //true если консоль включена
{
    if (TTY_BT == tty)
	return uartenable;
    else
	return true; //считаем что сериальная консоль всегда доступна
}


void ttyflush(uint8_t tty)
{
    if ((TTY_ALL == tty)||(TTY_BT == tty))
    {
	if (uartenable)
	    Serial.flush();
    }
    if ((TTY_ALL == tty)||(TTY_SERIAL == tty))
        ttySerial.flush();
}


// *****************************************************************************
// * ФУНКЦИИ НАСТРОЙКИ АППАРАТУРЫ
// *****************************************************************************


void uartsw(bool on) //включает/отключает аппаратный UART
{
    if (uartenable == on)
	return; //если статус не изменился ничего не далать
    if (on)
    {
	//поскольку скорость в режиме AT комманд отличается от
	//обычной скорости
	if (btcmdmode)
	    Serial.begin(BT_TTY_SPEED_AT); //(9600,SERIAL_8E1);
	else
	{
	    if (btmastermode)
		Serial.begin(BT_TTY_SPEED_MASTER);
	    else
		Serial.begin(BT_TTY_SPEED_SLAVE);
	}
	while (!Serial)
	{
	    ; // wait for serial port to connect. Needed for native USB
	}
	uartenable = true;
    }
    else
    {
	Serial.end();
	uartenable = false;
    }
}


void btPower(bool on) //переключает питание BT модуля
{
    if (btpoweron == on)
	return; //если статус не изменился ничего не далать
    if (on)
    {
	//востанавливаем KEY
	digitalWrite(PIN_BTKEY, btcmdmode);    // HС-05 режим AT выкл
	//включаем питание
	digitalWrite(PIN_BTPOWER, LOW);
	//включаем UART
	uartsw(true);
	btpoweron = true;
	btrxbytes = 0; //запускаем счетчик принятых байтов за сеанс
	btstarttime = millis(); //запомним время включения
    }
    else
    {
	if (digitalRead(PIN_BTLED) == HIGH)
	    btsessionok++; //счетчик успешных сессий
	else
	    btsessionerr++; //счетчик неуспешных сессий
	//завершаем заботу UART
	uartsw(false);
	//выходы контроллера идущие на BT в низкое состояние
	digitalWrite(PIN_BTRX, LOW);     //выход MK:TX->BT:RX
	digitalWrite(PIN_BTKEY, LOW);    //выход МК->вход HС-05 KEY
	//собственно отключение питания
	digitalWrite(PIN_BTPOWER, HIGH);
	btpoweron = false;
	state.btuptimems += millis()-btstarttime; //увеличить счетчик наработки BT
    }
}


void btAtCmdMode(bool on) //включить/выключить режим AT команд BT модуля 
{
    if (on == digitalRead(PIN_BTKEY)) //режим не изменился
	return;
    btcmdmode = on;
    //ХАРД РЕБУТ МОДУЛЯ
    btPower(false);
    delay(1000);
    btPower(true);
    delay(2000);
}


bool btCMode(uint8_t v) //1-коннектится ко всем подряд 0-только к определенному
{
    bool result = false;
    bool btcmdmodesave = btcmdmode;
    bool btpoweronsave = btpoweron;
    btAtCmdMode(true);
    if (v == 1)
    {
	if (!execAtCmd("+CMODE=1", 5))
	    result = false;
    }
    else
    {
	if (!execAtCmd("+CMODE=0", 5))
	    result = false;
    }
    btAtCmdMode(btcmdmodesave);
    btPower(btpoweronsave);
    return result;
}


bool btMasterMode(bool on) //включение/выключение master режима BT модуля
{
    bool result = false;
    if (!btcmdmode)
	btAtCmdMode(true);
    if (!execAtCmd("+ORGL", 5))
        goto errorexit;
    if (!execAtCmd("+RMAAD", 5))
        goto errorexit;
    if (on) //переключаем на master
    {
	if (!execAtCmd("+ROLE=1", 5))
	    goto errorexit;
	btmastermode = true;
    }
    else
    {
	if (!execAtCmd("+ROLE=0", 5))
	    goto errorexit;
	btmastermode = false;
    }
    if (!execAtCmd("+CMODE=1", 5))
        goto errorexit;
    if (!execAtCmd("+RESET", 5))
        goto errorexit;
    result = true;
    errorexit:
    btAtCmdMode(false);
    return result;
}


void sleepNow(int sleepmode) //отправляем МК в сон
{
    // выключаем ADC (он потребляет 250мкА)
    ADCSRA &= ~(1 << ADEN);
    //  The 5 different modes are:
    //      SLEEP_MODE_IDLE         -минимальная экономия (3мА)
    //      SLEEP_MODE_ADC
    //      SLEEP_MODE_PWR_SAVE
    //      SLEEP_MODE_STANDBY
    //      SLEEP_MODE_PWR_DOWN     -максимальная экономия (<1мкА)
    set_sleep_mode(sleepmode);   // sleep mode is set here

    sleep_enable();              // enables the sleep bit in the mcucr register
                                 // so sleep is possible. just a safety pin
    sleep_mode();                // here the device is actually put to sleep!!
                                 //
    sleep_disable();             // first thing after waking from sleep:
                                 // disable sleep...
}


long readVcc()
{
    ADCSRA |= (1 << ADEN); //turn ADC on
    // Read 1.1V reference against AVcc
    // set the reference to Vcc and the measurement to the internal 1.1V reference
    #if defined(__AVR_ATmega32U4__) || defined(__AVR_ATmega1280__) || defined(__AVR_ATmega2560__)
	ADMUX = _BV(REFS0) | _BV(MUX4) | _BV(MUX3) | _BV(MUX2) | _BV(MUX1);
    #elif defined (__AVR_ATtiny24__) || defined(__AVR_ATtiny44__) || defined(__AVR_ATtiny84__)
	ADMUX = _BV(MUX5) | _BV(MUX0);
    #elif defined (__AVR_ATtiny25__) || defined(__AVR_ATtiny45__) || defined(__AVR_ATtiny85__)
	ADMUX = _BV(MUX3) | _BV(MUX2);
    #else
	ADMUX = _BV(REFS0) | _BV(MUX3) | _BV(MUX2) | _BV(MUX1);
    #endif

    delay(75); // Wait for Vref to settle
    ADCSRA |= _BV(ADSC); // Start conversion
    while (bit_is_set(ADCSRA,ADSC)); // measuring

    uint8_t low  = ADCL; // must read ADCL first - it then locks ADCH  
    uint8_t high = ADCH; // unlocks both

    long result = (high<<8) | low;

    result = 1125300L / result; // Calculate Vcc (in mV); 1125300 = 1.1*1023*1000
    return result; // Vcc in millivolts
}


// *****************************************************************************
// * КОМАНДЫ ЯЗЫКА УПРАВЛЕНИЯ
// *****************************************************************************


void processCmdSAVEEEPROM(uint8_t tty)
{
    State state2; //копия чтобы не менялась в процессе записи
    state.cpuuptimems += millis() - uptimeresettime;
    uptimeresettime = millis();
    //if (btpoweron)
    //	state.btuptimems += (millis() - btstarttime);
    //запрещаем прерывания расчитываем CRC и делаем копию
    uint8_t oldSREG = SREG;
    cli();
    state.crc = crc8((uint8_t*)&state,sizeof(state)-1); // (-1 чтобы исключить поле CRC)
    memcpy(&state2, &state, sizeof(state));
    //разрешаем прерывания и записываем в eeprom из копии
    SREG=oldSREG;
    eeprom_update_block((const void*)&state2,(void*)&eepromstate, sizeof(state2));
}


void setCounter(uint8_t tty, uint8_t channel, uint8_t* svalue)
{
    //ttyserialprintf("\n\rsetCounter(%d,'%s')\n\r", channel, svalue);
    uint32_t v1 = 0; //целая часть
    uint32_t v2 = 0; //дробная
    uint8_t* p = (uint8_t*) strchr((char*)svalue, '.');
    if (p) //есть дробная часть
    {
	*p = '\0';
	p++; //теперь p указывает на первуюцифру дробной части
    }
    v1 = atol((char*)svalue); //целая
    if (channel == 0)
    {
	if (*p != '\0')
	    p[1] = '\0'; //оставляем только первую цифру после запятой
	v2 = atol((char*)p);
	//ttyserialprintf("\n\r v1=%lu v2=%lu \n\r", v1, v2);
	uint8_t oldSREG = SREG;
	cli();
	state.value[0] = v1 * 10 + v2;
	state.counter[0] = 0;
	SREG = oldSREG;
    }
    if ((channel == 1)||(channel == 2))
    {
	if (strlen((char*)p) > 3) //только три символа после запятой
	    p[3] = '\0';
	v2 = atol((char*)p);
	switch (strlen((char*)p))
	{
	    case 1:
		v2 *= 100;
		break;
	    case 2:
		v2 *= 10;
		break;
	    default:
		;//nothing
	}
	uint8_t oldSREG = SREG;
	cli();
	if (channel == 1)
	    state.value[1] = v1*1000 + v2;
	else
	    state.value[2] = v1*1000 + v2;
	SREG = oldSREG;
    }
    serialprintf(tty, "OK"ENDL);
}


void processCmdSET(uint8_t tty)
{
    uint8_t buf[MAX_TOKEN_SIZE];
    uint8_t* operand = getTokenFromTtyBuf((uint8_t*)&buf, MAX_TOKEN_SIZE); //прочитать операнд
    if (strcmp((char*)operand,"") == 0)
    {
	serialprintf(tty, "ERR: too few operands"ENDL);
	return;
    }
    //управление светодиодом
    if (strcmp("LED",(char*)operand) == 0)
    {
        operand = getTokenFromTtyBuf((uint8_t*)&buf, MAX_TOKEN_SIZE); //прочитать операнд
	if (strcmp("ON",(char*)operand) == 0)
	{
	    serialprintf(tty, "OK LED set on"ENDL);
	    ledenable = true;
	    goto successexit;
	}
	if (strcmp("OFF",(char*)operand) == 0)
	{
	    serialprintf(tty, "OK LED set off"ENDL);
	    ledenable = false;
	    goto successexit;
	}
	goto errorexit;
    }
    //управление счетчиками (#0 #1 #2)
    if (strlen((char*)operand) == 2)
    {
	uint8_t nchannel = 0;
	if (operand[0] == '#')
	{
	    //номер счетчика
	    nchannel = atoi((char*)&operand[1]);
	    operand = getTokenFromTtyBuf((uint8_t*)&buf, MAX_TOKEN_SIZE); //прочитать операнд
	    if (nchannel > 2)
		serialprintf(tty, "ERR: invalid channel %d"ENDL, nchannel);
	    else
		setCounter(tty, nchannel, operand);
	    goto successexit;
	}
    }
    //управление интервалами передачи показаний
    if ((strlen((char*)operand) == 4)&&(strncmp("GAP",(char*)operand,3) == 0))
    {
	uint8_t nchannel = atoi((char*)&operand[3]);
	if (nchannel < 3)
	{
	    operand = getTokenFromTtyBuf((uint8_t*)&buf, MAX_TOKEN_SIZE); //прочитать операнд
	    uint8_t value = atoi((char*)operand);
	    state.gap[nchannel] = value;
	    serialprintf(tty,"GAP%d=%d"ENDL,nchannel,value);
	    goto successexit;
	}
    }
    //установить число импульсов на 0.1 KWH
    if (strcmp("IMP01KWH",(char*)operand) == 0)
    {
	operand = getTokenFromTtyBuf((uint8_t*)&buf, MAX_TOKEN_SIZE); //прочитать операнд
	uint16_t value = atoi((char*)operand);
	state.imp01KWH = value;
	serialprintf(tty,"IMP01KWH=%d"ENDL,value);
	goto successexit;
    }
    //управление BT модулем
    if (strcmp("BT",(char*)operand) == 0)
    {
        operand = getTokenFromTtyBuf((uint8_t*)&buf, MAX_TOKEN_SIZE); //прочитать операнд
	if (strcmp("ON",(char*)operand) == 0)
	{
	    ttyflush(tty); //иначе мусор на консоле
	    btPower(true);
	    serialprintf(tty, "BT on"ENDL);
	    goto successexit;
	}
	if (strcmp("OFF",(char*)operand) == 0)
	{
	    btPower(false);
	    serialprintf(tty, "BT off"ENDL);
	    goto successexit;
	}
	if (strcmp("ATCMDMODE",(char*)operand) == 0)
	{
	    operand = getTokenFromTtyBuf((uint8_t*)&buf, MAX_TOKEN_SIZE); //прочитать операнд
	    if (strcmp("ON",(char*)operand) == 0)
	    {
		serialprintf(tty, "AT MODE ON"ENDL);
		btAtCmdMode(true);
		goto successexit;
	    }
	    if (strcmp("OFF",(char*)operand) == 0)
	    {
		serialprintf(tty, "AT MODE OFF"ENDL);
		btAtCmdMode(false);
		goto successexit;
	    }
	    goto errorexit;
	}
	if (strcmp("MASTERMODE",(char*)operand) == 0)
	{
	    if ((!btcmdmode)||(!btpoweron))
	    {
		serialprintf(tty,"ERR: BT off or not at AT mode"ENDL);
		goto successexit;
	    }
	    operand = getTokenFromTtyBuf((uint8_t*)&buf, MAX_TOKEN_SIZE); //прочитать операнд
	    if (strcmp("ON",(char*)operand) == 0)
	    {
		serialprintf(tty, "MASTERMODE ON"ENDL);
		btMasterMode(true);
		goto successexit;
	    }
	    if (strcmp("OFF",(char*)operand) == 0)
	    {
		serialprintf(tty, "MASTERMODE OFF"ENDL);
		btMasterMode(false);
		goto successexit;
	    }
	    goto errorexit;
	}
/*
	if (strcmp("RADDR",(char*)operand) == 0)
	{
	    operand = getTokenFromTtyBuf((uint8_t*)&buf, MAX_TOKEN_SIZE); //прочитать операнд
	    if (strlen((char*)operand) > 8)
	    {
		strncpy(btraddr,(char*)operand,sizeof(btraddr)-1);
		btraddr[sizeof(btraddr)-1] = '\0';
		serialprintf(bt,"BT RADDR='%s'"ENDL, btraddr);
		goto successexit;
	    }
	}
*/
	goto errorexit;
    }
    errorexit:
    serialprintf(tty, "ERR: invalid parameter '%s'"ENDL, operand);
    successexit:
    return;
}


void processCmdGETCounterAll(uint8_t tty)
{
    long vcc = readVcc(); //напряжение питания в мВ
    unsigned long cputime = state.cpuuptimems/1000 + (millis() - uptimeresettime)/1000;
    unsigned long bttime  = state.btuptimems/1000;
    if (btpoweron)
	bttime += (millis() - btstarttime)/1000;
    serialprintf(tty, "#0:%05lu.%01u(kwh) %05u #1:%05lu.%03u(m3) #2:%05lu.%03u(m3) CPUtime:%lu(s) BTtime:%lu(s) Vbat:%lu(mv) %s"ENDL
                    , v[0]/10, (unsigned int)(v[0]%10), c0 ,v[1]/1000, (unsigned int)(v[1]%1000), v[2]/1000, (unsigned int)(v[2]%1000)
                    , cputime, bttime, vcc, (vcc < LOWBAT_LEVEL)?"LOWBAT":""
                );
    serialprintf(tty, "FreeRam:[%d]"ENDL, freeRam());
    serialprintf(tty, "Sess:%lu/%lu"ENDL, btsessionok, btsessionerr);
}


void processCmdGetBuildInfo(uint8_t tty)
{
    serialprintf(tty, "VER:%s %s"ENDL,XSTR(BUILDVER),XSTR(BUILDTIME));
}


void processCmdGET(uint8_t tty)
{
    uint8_t buf[MAX_TOKEN_SIZE];
    uint8_t* operand = getTokenFromTtyBuf((uint8_t*)&buf, MAX_TOKEN_SIZE); //прочитать операнд
    //получить счетчики
    if (strcmp("COUNTER",(char*)operand) == 0)
    {
        operand = getTokenFromTtyBuf((uint8_t*)&buf, MAX_TOKEN_SIZE); //прочитать операнд
	if (strcmp("ALL",(char*)operand) == 0)
	{
	    processCmdGETCounterAll(tty);
	    goto successexit;
	}
    }
    //получить версию сборки
    if (strcmp("BUILD",(char*)operand) == 0)
    {
        operand = getTokenFromTtyBuf((uint8_t*)&buf, MAX_TOKEN_SIZE); //прочитать операнд
	if (strcmp("INFO",(char*)operand) == 0)
	{
	    processCmdGetBuildInfo(tty);
	    goto successexit;
	}
    }
    //получить LED статус
    if (strcmp("LED",(char*)operand) == 0)
    {
        operand = getTokenFromTtyBuf((uint8_t*)&buf, MAX_TOKEN_SIZE); //прочитать операнд
	if (strcmp("STATUS",(char*)operand) == 0)
	{
	    serialprintf(tty, "OK LED is %s"ENDL, (ledenable)?"enable":"disable");
	    goto successexit;
	}
    }
    //получить BT статус
    if (strcmp("BT",(char*)operand) == 0)
    {
        operand = getTokenFromTtyBuf((uint8_t*)&buf, MAX_TOKEN_SIZE); //прочитать операнд
	if (strcmp("STATUS",(char*)operand) == 0)
	{
	    serialprintf(tty
                        , "OK BLUETOOTH POWER:%s ATCMDMODE:%s MASTER:%s LINK:%s"ENDL
                        , (digitalRead(PIN_BTPOWER))?"OFF":"ON"
                        , (digitalRead(PIN_BTKEY))?"ON":"OFF"
                        , (btmastermode)?"YES":"NO"
                        , (digitalRead(PIN_BTLED))?"CONNECTED":"DISCONNECTED"
                        );
	    goto successexit;
	}
    }
    //получить интервалы передачи показаний
    if ((strlen((char*)operand) == 4)&&(strncmp("GAP",(char*)operand,3) == 0))
    {
	uint8_t nchannel = atoi((char*)&operand[3]);
	if (nchannel < 3)
	{
	    serialprintf(tty,"GAP%d=%d"ENDL,nchannel,state.gap[nchannel]);
	    goto successexit;
	}
    }
    //получить число импульсов на 0.1 КВТЧАС
    if (strcmp("IMP01KWH",(char*)operand) == 0)
    {
	serialprintf(tty,"IMP01KWH=%d"ENDL,state.imp01KWH);
	goto successexit;
    }
    if (strcmp((char*)operand,"") == 0)
	serialprintf(tty, "ERR: too few operands"ENDL);
    else
	serialprintf(tty, "ERR: invalid parameter '%s'"ENDL, operand);
    successexit:
    return;
}


bool execAtCmd(const char* cmd, uint16_t timeoutsec) //диалог с BT модулем
{
    char buf[128];
    uint8_t n = 0;
    serialprintf(TTY_SERIAL,"execAtCmd(%s)"ENDL, cmd);
    serialprintf(TTY_BT, "AT%s"ENDL,cmd); //оптправить команду
    unsigned long time1 = millis();
    bool buffull = false;
    //ждать результат
    while(true)
    {
	if (ttyavailable(TTY_BT) > 0)
	{
	    uint8_t ch = ttyread(TTY_BT);
	    time1 = millis();
	    serialprintf(TTY_SERIAL, "%c", ch);
	    //пришел конец строки
	    if ((ch == '\n')||(ch == '\r'))
	    {
		if (strncmp(buf,"ERROR:",strlen("ERROR:")) == 0)
		    timeoutsec = 1; //уменьшить таймаут
		if (strncmp(buf,"FAIL",strlen("FAIL")) == 0)
		    timeoutsec = 5; //уменьшить таймаут
		if (n>=2)
		    if (strncmp(&buf[n-2],"OK",2) == 0)
		    timeoutsec = 1; //уменьшить таймаут
	    }
	    else
	    {
		if (!buffull)
		    buf[n++] = ch;
	    }
	    //переполнился буфер
	    if (n >= sizeof(buf) - 1)
	    {
		buffull = true;
	    }
	}
	else
	{
	    //if (done)
	    //	break; //входных символов больше нет и ОК уже получен
	    //таймаут
	    unsigned long timenow = millis();
	    if ((timenow - time1)/1000 > timeoutsec)
	    {
		//serialprintf(false,"TIMEOUT=%us"ENDL,timeoutsec);
		break;
	    }
	}
    }
    buf[n] = '\0';
    serialprintf(TTY_SERIAL,"result:[%s] size=%d"ENDL, buf, n);
    return true;
}


void processCmdAT(uint8_t tty, char* cmd) //AT
{
    if (btcmdmode)
	execAtCmd(cmd, 60);
    else
	serialprintf(tty,/*"BT unit not at AT commands mode. Try SET BT ATCMDMODE ON first"*/ENDL);
}


void processTtyCmd(uint8_t tty) //парсить строку команды
{
    uint8_t* buf[MAX_TOKEN_SIZE];
    uint8_t* token = getTokenFromTtyBuf((uint8_t*)&buf, MAX_TOKEN_SIZE); //прочитать первый токен (команды)
    if (strcmp("",(char*)token) == 0) //пустая строка
    {
	goto successexit;
    }
    if (strcmp("SET",(char*)token) == 0)
    {
	processCmdSET(tty);
	goto successexit;
    }
    if (strcmp("GET",(char*)token) == 0)
    {
	processCmdGET(tty);
	goto successexit;
    }
    if (strncmp("AT",(char*)token, 2) == 0)
    {
	processCmdAT(tty,(char*)token);
	goto successexit;
    }
    if (strcmp("SAVEEEPROM",(char*)token) == 0)
    {
	processCmdSAVEEEPROM(tty);
	goto successexit;
    }
    if (strcmp("RESETUPTIME",(char*)token) == 0)
    {
	uptimeresettime = millis();
	state.cpuuptimems = 0;
	state.btuptimems = 0;
	serialprintf(tty,"OK"ENDL);
	goto successexit;
    }
    if (strcmp("REBOOT",(char*)token) == 0)
    {
	//state.cpuuptimems += millis() - uptimeresettime; //добавить текущую наработку
	processCmdSAVEEEPROM(tty); //сохраняем состояние в EEPROM
	resetFunc(); //вызываем программный ребут
	//сюда мы попасть по идее не должны
	goto successexit;
    }
    if (strncmp("QUIT",(char*)token, 2) == 0)
    {
	if (TTY_BT == tty)
	{
	    serialprintf(tty,"byebye :)"ENDL);
	    delay(100);
	    btPower(false);
	}
	{
	    serialprintf(tty,"For BTtty only"ENDL);
	}
	goto successexit;
    }
    serialprintf(tty, "ERR: unkcmd '%s'"ENDL, token);
    successexit:
    ttyBufClear(); //очистить буфер (от лишних операндов и ошибочных комманд)
    return;
}


int processTtySerialInput(uint8_t tty) //обработать приходящие на порт символы (вернет кол-во пришедших символов)
{
    int result = 0;
    if (!ttyenable(tty))
	return result; //консоль выключена
    bool needflush = false;
    //принять символ с tty консоли
    while(ttyavailable(tty) > 0)
    {
	needflush = true;
	uint8_t ch = ttyread(tty);
	result++;
	switch(ch)
	{
	    case 8: //backspace
		ttyBufPopBack();
		serialprintf(tty,"%c", ch); //отправить эхо обратно
		break;
	    case '\r': //просто игнорируем
		break;
	    case '\n':
		serialprintf(tty, "%s", ENDL); //перевести строку
		ttyBufPush('\0');
		processTtyCmd(tty); //если пришел конец строки то парсим команду
		break;
	    default:
		serialprintf(tty,"%c", ch); //отправить эхо обратно
		ttyBufPush(toupper(ch)); //вставить в буфер
	} //switch
    }
    if (needflush)
    {
	ttyflush(tty);
	delay(1);
    }
    return result;
}


// *****************************************************************************
// * 
// *****************************************************************************


void setup() //первоначальная установка при старте
{
    ttySerial.begin(9600); //включаем софтовый uart для отладочной консоли
    processCmdGetBuildInfo(TTY_ALL);
    pinMode(PIN_LED, OUTPUT);      // sets the digital pin as output
    pinMode(PIN_PHOTOR, INPUT);    // вход для фоторезистора
    digitalWrite(PIN_PHOTOR, LOW); // выключить подтягивающий резистор 20КОм для входа
    pinMode(PIN_GERKON1, INPUT/*_PULLUP*/);    // вход для геркона1
    //digitalWrite(PIN_GERKON1, HIGH);
    pinMode(PIN_GERKON2, INPUT/*_PULLUP*/);    // вход для геркона2
    digitalWrite(PIN_GERKON2, LOW); // выключить подтягивающий резистор 20КОм для входа
    //инициализировать выводы управления BT
    //линии RX/TX инициализируются в Serial.begin()
    pinMode(PIN_BTKEY, OUTPUT);                // выход на линию KEY BT модуля
    //digitalWrite(btkeypin, LOW);              // РС-05 режим AT выкл
    pinMode(PIN_BTPOWER, OUTPUT);              // выход на управление пританием BT модуля
    //digitalWrite(btpowerpin, HIGH);           // РС-05 питание вкл
    pinMode(PIN_BTLED, INPUT);                  // вход для сигнала LED от BT модуля
    //прочитать данные из EEPROM (от прошлого перезапуска)
    eeprom_read_block((void*)&state, (const void*)&eepromstate, sizeof(state));
    if ( state.crc != crc8((uint8_t*)&state,sizeof(state)-1)) //проверяем CRC (-1 чтобы исключить поле CRC)
    {
	serialprintf(TTY_ALL, "EEPROM CRC ERR"ENDL);
	//ставим значения по умолчанию
	memset(&state, 0, sizeof(state)); //сбросить счетчики если CRC не совпала
	state.imp01KWH = 1280; //импульсов на 0.1 KWH
	state.gap[0]   = 10;   //передавать каждые 1 kwh
	state.gap[1]   = 100;  //передавать каждые 100l
	state.gap[2]   = 100;  //передавать каждые 100l
    }
    //мигнуть диодом 3 раза
    for (uint8_t i = 0; i < 3; i++)
    {
	digitalWrite(PIN_LED, HIGH);   // sets the LED on
	delay(500);
	digitalWrite(PIN_LED, LOW);    // sets the LED off
	delay(500);
    }
    attachInterrupt(0, inthandlerPhoto, FALLING/*RISING*/); //включить обработчик прерывания LOW->HIGH
    attachInterrupt(1, inthandlerGerkon1, CHANGE); //включить обработчик прерывания первого геркона

//    pciSetup(gerkon2pin); //обработчик второго геркона
//    pciSetup(rxsoftttypin); //обработчик RX softtty пина

    //разрешаем PIN CHANGE прерывания
    PCICR  = (1<<PCIE0)|(1<<PCIE1)|(1<<PCIE2); //разрешить прерывание PCINT0 PCINT1 PCINT2
    PCMSK0 = (1<<PCINT0); // выбрать входы для PCINT0_vect (вывод dip8)
    PCMSK1 = (1<<PCINT8); //(1<<PCINT9)|(1<<PCINT8); //выбрать входы для PCINT1_vect (PCINT8 - dip23)
    PCMSK2 = (1<<PCINT16); // выбрать входы для PCINT2_vect (PCINT16/RXD - dip2)
    //перевключаем BT модуль
    btpoweron = true;
    btPower(false); //выключаем питание BT
    delay(2000);
    btPower(true); //включаем питание BT

    uint8_t sec = 0;
    //ждем когда приконнектится BT но не более 60 сек
    while(!digitalRead(PIN_BTLED))
    {
	delay(1000);
	sec++;
	if( sec > 60)
	    break;
    }
    sec = 0;
    //если коннекта нет, то делаем сброс настроек BT контроллера
    //и пытаемся приконнектиться сначала в режиме мастер (на PC будет запрос пина) и, если и это
    //не получилось, то переходим в режим слейв
    if (!digitalRead(PIN_BTLED))
    {
	btMasterMode(true);
	//ждем когда приконнектится но не более 60 сек
	while(!digitalRead(PIN_BTLED))
	{
	    delay(1000);
	    sec++;
	    if( sec > 60)
	    {
		btMasterMode(false); //переходим в слейв режим
		serialprintf(TTY_SERIAL, "BT FALLBACK TO SLAVE"ENDL);
		break;
	    }
	}
    }
    serialprintf(TTY_ALL, ENDL"Free SRAM %db"ENDL,freeRam());
    ttyflush(TTY_ALL);
    delay(500);
}




void loop()
{
    static uint32_t v0bak = 0;
    static uint32_t v1bak = 0;
    static uint32_t v2bak = 0;
    static bool needbtsend = false;
    static bool needttysend = false;
    static unsigned long lastbtsymboltime = 0;

    //делаем копию переменных при запрещенных прерываниях
    uint8_t oldSREG = SREG;
    cli();
    memcpy(v,state.value, sizeof(state.value));
    c0 = state.counter[0];
    SREG = oldSREG; //востанавливаем значение прерываний
    if (ledenable)
	digitalWrite(PIN_LED, HIGH);   // sets the LED on
    //костыль
    if (ttywakeup) //проснулись от tty символа
    {
	delay(60);
	//сбросить флаги пробуждения от консоли
	oldSREG = SREG;
	cli();
	ttywakeup = false;
	SREG = oldSREG;
    }
    //отображаем если счетчики изменились на величину больше пороговой
    if ((v[0] - v0bak >= state.gap[0])||(v[1] - v1bak >= state.gap[1])||(v[2] - v2bak >= state.gap[2]))
    {
	needbtsend = true;
	needttysend = true;
	btPower(true); //включаем питание BT
    }
    //передать показания на TTY если нужно
    if (needttysend)
    {
	processCmdGETCounterAll(TTY_SERIAL); //передать показания счетчиков на TTY
	ttyflush(TTY_SERIAL);
	needttysend = false;
	//запомнить переданные показания
	v0bak = v[0];
	v1bak = v[1];
	v2bak = v[2];
    }
    //передать показания на BT если есть соединение
    if ((needbtsend)&&(digitalRead(PIN_BTLED) == HIGH)) //есть данные для передачи на BT и есть BT соединение
    {
	if (!btcmdmode)
	{
	    processCmdGETCounterAll(TTY_BT); //передать показания счетчиков на BT 
	    ttyflush(TTY_BT);
	    needbtsend = false;
	    lastbtsymboltime = millis(); //время отправки последнего символа
	    processCmdSAVEEEPROM(TTY_BT);
	}
    }
    //проверить буфер bt на наличие символов
    if (!btcmdmode)
    {
	if (processTtySerialInput(TTY_BT) > 0)
	{
	    lastbtsymboltime = millis(); //время прихода последнего символа
	}
    }
    //проверить буфер serial tty на наличие символов
    processTtySerialInput(TTY_SERIAL);
    digitalWrite(PIN_LED, LOW);    // sets the LED off
    //отключаем питание BT при определенных условиях
    if (
	//исходящих нет коннект  есть интерактивный сеанс и нет входящих символов более 300сек
	((!needbtsend) && (digitalRead(PIN_BTLED)) /*&& (btrxbytes >19)*/ && (millis()-lastbtsymboltime > 300000)) ||
	//исходящих нет коннект есть НЕинтерактивный сеанс и нет входящих символов более 5сек
//	((!needbtsend) && (digitalRead(PIN_BTLED)) /*&& (btrxbytes <=19)*/ && (millis()-lastbtsymboltime > 5000)) ||
	//мастеррежим, исходящие символы есть, коннекта нет и слишком долго не удается соединиться с BT
        ((btmastermode) && (needbtsend) && (!digitalRead(PIN_BTLED)) && ((millis()-btstarttime) > 30000))
       )
    {
	needbtsend = false; //эти показания переданы не будут (и хрен с ними)
	btPower(false); //отключаем BT для экономии батареи
    }
    //если BT выключен и уже набралось несколько успешных сессий переключаемся на AT+CMODE=0
    if ((!btcmodezero)&&(!btpoweron)&&(btsessionok == 5))
    {
	btCMode(0);
	btcmodezero = true;
	serialprintf(TTY_SERIAL,"BT->CMODE=0"ENDL);
    }
    //уходим в сон
    oldSREG = SREG;
    cli();
    sleeping = true;
    SREG = oldSREG; //востанавливаем значение прерываний
    sleepNow((btpoweron)?SLEEP_MODE_IDLE:SLEEP_MODE_PWR_DOWN); // sleep function called here
    //проснулись
    oldSREG = SREG;
    cli();
    sleeping = false;
    SREG = oldSREG; //востанавливаем значение прерываний
}
